// maincontroller code phaeno pinball
// runs on arduino mega 2560


//INPUTS:
#define I_PLUNGER 9
#define I_FLIPPER_L 12
#define I_FLIPPER_R 11
#define I_FLIPPER_O 10
#define I_LIGHTS_L 6
#define I_LIGHTS_R 5
#define SCORE A6
#define IR_PLUNGER A7

//OUTPUTS:
#define O_FLIPPER_L A0
#define O_FLIPPER_R A1
#define O_FLIPPER_O A2
#define O_PLUNGER A3
#define O_LIGHTS_L A4
#define O_LIGHTS_R A5
#define CALC_AC 14
#define CALC_0 15
#define CALC_1 16
#define CALC_5 17
#define CALC_PLUS 18
#define CALC_EQUALS 19
#define BELL 21
#define BUZZER 20
#define TIE_LIGHT A14
#define BONUS_LIGHT A15

#define RINGBUFFER_SIZE 30       /*take it easy alarm: amount of maximum button strokes...*/
#define TIMEFRAME  10000L        /*take it easy alarm: ...in this timeframe (in millis) causes alarm*/

//ringbuffer variables for take it easy alarm and tilt:

long flipper_ringbuffer[RINGBUFFER_SIZE];    //ringbuffer which saves the times when flipper fingers are released 
int  flipper_ringbuffer_count=0; //(in order to know when user should take it easy)        
long take_it_easy=0;             //variable saves "take it easy" - condition. 0=everything ok / >0=take it easy alarm
boolean tilt= false;             //tilt=true aborts the game when player does not react to take it easy alarm within 5 seconds
  
//flipper finger variables for debounce

boolean flipper_l=LOW;    //stores flipper button input
boolean flipper_r=LOW;
boolean flipper_o=LOW;

boolean p_flipper_l=LOW; //last flipper button position
boolean p_flipper_r=LOW;
boolean p_flipper_o=LOW;

long    t_flipper_l;     //time when flipper finger was released (for debounce)
long    t_flipper_r;
long    t_flipper_o;

int score=0;             //keeps track of score
long t_score_off=0;      //timer variable will make sure that score counting on calculator ends again
boolean set_score=false; //stores info that player just scored (for switching off '=' on calculator and light effect relays)

//ball out
int ball_out=0; //will count up when ball IR sensor says ball is out
int light_chain_counter=0; //counts light chain effects
 
void setup(){//====================================================================== SETUP
               

pinMode( I_PLUNGER   ,INPUT_PULLUP);
pinMode( I_FLIPPER_L ,INPUT_PULLUP);
pinMode( I_FLIPPER_R ,INPUT_PULLUP);
pinMode( I_FLIPPER_O ,INPUT_PULLUP);
pinMode( I_LIGHTS_L  ,INPUT_PULLUP);
pinMode( I_LIGHTS_R  ,INPUT_PULLUP);

pinMode( O_FLIPPER_L ,OUTPUT);
pinMode( O_FLIPPER_R ,OUTPUT);
pinMode( O_FLIPPER_O ,OUTPUT);
pinMode( O_PLUNGER   ,OUTPUT);
pinMode( O_LIGHTS_L  ,OUTPUT);
pinMode( O_LIGHTS_R  ,OUTPUT);
pinMode( CALC_AC     ,OUTPUT);
pinMode( CALC_0      ,OUTPUT);
pinMode( CALC_1      ,OUTPUT);
pinMode( CALC_5      ,OUTPUT);
pinMode( CALC_PLUS   ,OUTPUT);
pinMode( CALC_EQUALS ,OUTPUT);
pinMode( BELL        ,OUTPUT);
pinMode( BUZZER      ,OUTPUT);
pinMode( TIE_LIGHT   ,OUTPUT);
pinMode( BONUS_LIGHT ,OUTPUT);
  


Serial.begin(9600);
Serial.println(millis());

//fill ringbuffer
for (int i=0;i<RINGBUFFER_SIZE;i++){
flipper_ringbuffer[i]=-TIMEFRAME;  
  }
}

void loop(){//======================================================================= LOOP
  
  
  //score
  
  if(analogRead(SCORE)<500 && t_score_off<millis()){
    add_score();
    t_score_off=millis()+500;
    set_score=true;
    Serial.println(score);
  }
  if (t_score_off<millis()&&set_score){
    digitalWrite(CALC_EQUALS,LOW);
    digitalWrite(O_LIGHTS_L,LOW);
    digitalWrite(O_LIGHTS_R,LOW);
    set_score=false;
  }

  
  //check light chain buttons (only if the effects relays are not operated by the scoring algorithm)
  if (set_score==false){check_light_chain_buttons();}
  
  //check flipper buttons
  check_flipper_buttons();
  
  //move flipper fingers
  move_flipper_fingers();
  
  //take it easy alarm?
  if (take_it_easy){
    take_it_easy_on(); //switch alarm on
    if (flipper_ringbuffer[(flipper_ringbuffer_count+1)%RINGBUFFER_SIZE]+TIMEFRAME<millis()){ // is old ringbuffer time within timeframe? switch the take it easy alarm off again
      take_it_easy=0;
    }
  }else{
    take_it_easy_off(); //switch alarm off when player behaves again for a while
  }
  
  //tilt?
  if (millis()-take_it_easy>10000L && take_it_easy>0)tilt=true; //if player does not react to take it easy alarm -> tilt the machine!
  if (tilt) {           //machine is tilted!
    Serial.println("------> T/I/L/T/!");
    take_it_easy_off(); //switch alarm off when machine is tilted
    digitalWrite(O_FLIPPER_L,LOW);//release flipper fingers
    digitalWrite(O_FLIPPER_R,LOW);
    digitalWrite(O_FLIPPER_O,LOW);
    digitalWrite(BELL,HIGH);
    delay(50);
    digitalWrite(BELL,LOW);
    restart_game();
  }
  
  //ball out?
  if (analogRead(IR_PLUNGER)>230){                  // check infrared sensor
    ball_out++;
  }else{
    ball_out=0;
  }
  if (ball_out==100){                               // plunger IR sensor recognized ball 100x in a row
    digitalWrite(O_FLIPPER_L,LOW);                  // release flipper fingers
    digitalWrite(O_FLIPPER_R,LOW);
    digitalWrite(O_FLIPPER_O,LOW);
    Serial.println("Ball is out - game over");
    take_it_easy_off();
    add_bonus(score/10+1);                          // calculate & add bonus
    restart_game();
  }
}



//=================================================================================== FUNCTIONS

void check_light_chain_buttons(){//-------------------------check rear buttons to switch light chain effects manually
  if (digitalRead(I_LIGHTS_L)==LOW){digitalWrite(O_LIGHTS_L,HIGH);}else{digitalWrite(O_LIGHTS_L,LOW);}
  if (digitalRead(I_LIGHTS_R)==LOW){digitalWrite(O_LIGHTS_R,HIGH);}else{digitalWrite(O_LIGHTS_R,LOW);}
}

void restart_game(){//-------------------------start sequence
  Serial.println("restart game");
  digitalWrite(CALC_EQUALS,LOW); 
  score=0; //reset score
  tilt=false;
  
  Serial.println("reset light chain effect");
  while(light_chain_counter!=0){ //reset light chain effect
    digitalWrite(O_LIGHTS_L,HIGH);
    digitalWrite(O_LIGHTS_R,HIGH);
    delay(100);
    digitalWrite(O_LIGHTS_L,LOW);
    digitalWrite(O_LIGHTS_R,LOW);
    delay(100);
    light_chain_counter=(light_chain_counter+1)%8;  
  }
  
  // make sure ball is inside plunger (and wait until it is)
  ball_out=0;
  while (ball_out<100){                                // wait for ball in plunger
     check_light_chain_buttons();                      // check light chain buttons
     if (analogRead(IR_PLUNGER)>230){                  // check infrared sensor
      ball_out++;                                      // ball in plunger detected -> increment ball_out     
    }else{
      ball_out=0;
    }
  }
  
  
  // ball is inside plunger!
  Serial.println("ball in plunger detected");
  while(digitalRead(I_PLUNGER)){                // wait until plunger button is pushed
    check_light_chain_buttons();
  }                            
  Serial.println("plunger button is pushed!");
  Serial.println("initialize calculator...");
  int d=75;
  digitalWrite(O_PLUNGER,HIGH);                 // shoot in new ball (start hairdryer)
  digitalWrite(BELL,HIGH);
  digitalWrite(CALC_AC,HIGH);delay(d);          // switch calculator on
  digitalWrite(BELL,LOW);
  digitalWrite(CALC_AC,LOW);delay(d);
  digitalWrite(CALC_1,HIGH);delay(d);           // enter 150 to calculator
  digitalWrite(CALC_1,LOW);delay(d);
  digitalWrite(CALC_5,HIGH);delay(d);
  digitalWrite(CALC_5,LOW);delay(d);
  digitalWrite(CALC_0,HIGH);delay(d);
  digitalWrite(CALC_0,LOW);delay(d);
  digitalWrite(CALC_PLUS,HIGH);delay(d);        // push + on calculator
  digitalWrite(CALC_PLUS,LOW);delay(d);
  digitalWrite(CALC_EQUALS,HIGH);delay(d);      // push = on calculator
  digitalWrite(CALC_EQUALS,LOW);delay(d);
  for (int i=0;i<4000;i++){                     // plunger hairdryer is kept on for a while -- flipper fingers react to buttons
    delay(1);
    check_flipper_buttons();
    move_flipper_fingers();
  }
  digitalWrite(O_PLUNGER,LOW);                  // switch off plunger/hairdryer
}



void check_flipper_buttons(){//-------------------------check flipper buttons  
  
  if (digitalRead(I_FLIPPER_L)==LOW){ //left flipper button
    flipper_l=HIGH;
  }else{
    if (flipper_l==HIGH)t_flipper_l=millis(); //save time when flipper button was released
    flipper_l=LOW;
  }
  
  if (digitalRead(I_FLIPPER_R)==LOW){ // right flipper button
    flipper_r=HIGH;
  }else{
    if (flipper_r==HIGH)t_flipper_r=millis();//save time when flipper button was released
    flipper_r=LOW;
  }
  
  if (digitalRead(I_FLIPPER_O)==LOW){ //upper flipper button
    flipper_o=HIGH;
  }else{
    if (flipper_o==HIGH)t_flipper_o=millis();//save time when flipper button was released
    flipper_o=LOW;
  }
}

void move_flipper_fingers(){//-------------------------move flipper fingers
  if (flipper_l==HIGH){
    digitalWrite(O_FLIPPER_L,HIGH);               //move flipper finger
    p_flipper_l=HIGH;                             //save current state as previous state for upcoming loop cycle
  }
  if (flipper_l==LOW && t_flipper_l+25<millis()){ //release flipper finger when button was released for 25ms (debounce)
    digitalWrite(O_FLIPPER_L,LOW);                //release flipper finger
    if (p_flipper_l==HIGH){save_flipper_time();}  //save time of release in ringbuffer for take it easy alarm when necessary
    p_flipper_l=LOW;                              //save current state as previous state for upcoming loop cycle
  }
  //                                              --> do the same thing with right flipper finger
  if (flipper_r==HIGH){
    digitalWrite(O_FLIPPER_R,HIGH);
    p_flipper_r=HIGH;
  }
  if (flipper_r==LOW && t_flipper_r+25<millis()){
    digitalWrite(O_FLIPPER_R,LOW);
    if (p_flipper_r==HIGH){save_flipper_time();}
    p_flipper_r=LOW;
  }
  //                                              --> do the same thing with upper flipper finger
  if (flipper_o==HIGH){
    digitalWrite(O_FLIPPER_O,HIGH);
    p_flipper_o=HIGH;
  }
  if (flipper_o==LOW && t_flipper_o+25<millis()){
    digitalWrite(O_FLIPPER_O,LOW);
    if (p_flipper_o==HIGH){save_flipper_time();}
    p_flipper_o=LOW;
  }
}

void save_flipper_time(){//------------------------- write the current time in a ringbuffer in order to find out how fast/often flippers are used

  flipper_ringbuffer_count++;                                         // update ringbuffer index
  flipper_ringbuffer_count=flipper_ringbuffer_count%RINGBUFFER_SIZE;  // 
  flipper_ringbuffer[flipper_ringbuffer_count]=millis();              // save current time in ringbuffer
  if ((flipper_ringbuffer[(flipper_ringbuffer_count+1)%RINGBUFFER_SIZE]+TIMEFRAME)>millis()){ // is old ringbuffer time out of timeframe?  
   if (take_it_easy==0) take_it_easy=millis();                                                // yes: then take it easy and save time when take it easy alarm went on first
   Serial.print("TAKE IT EASY ALARM ON AT TIME:");
   Serial.println(take_it_easy);
  }

}


void add_score(){//------------------------------- add score on calculator
  digitalWrite(CALC_EQUALS,HIGH);       // =
  score++;
  if(score%10==1){                      //change light effects on light chains every ten points
  
    digitalWrite(O_LIGHTS_L,HIGH);
    digitalWrite(O_LIGHTS_R,HIGH);
    light_chain_counter=(light_chain_counter+1)%8;
    Serial.print("::::::::::::light chain effect: ");
    Serial.println(light_chain_counter);
  }
  Serial.println("SCORE++!");
}

void add_bonus(int bonus_score){//---------------- add bonus on calculator
  Serial.println("bonus time!");
  digitalWrite(BONUS_LIGHT,HIGH);
  int d=120;
  digitalWrite(CALC_PLUS,HIGH);delay(d);        // +
  digitalWrite(CALC_PLUS,LOW);delay(d);
  digitalWrite(CALC_1,HIGH);delay(d);           // enter 10
  digitalWrite(CALC_1,LOW);delay(d);
  digitalWrite(CALC_0,HIGH);delay(d);
  digitalWrite(CALC_0,LOW);delay(d);
  
  
  for (int i=0;i<bonus_score;i++){              //add ten points per bonus and ring bell
    digitalWrite(BONUS_LIGHT,HIGH);
    digitalWrite(CALC_EQUALS,HIGH);             // =
    digitalWrite(BELL,HIGH);
    delay(d/2);    
    digitalWrite(BELL,LOW);  
    delay(d/2); 
    digitalWrite(BELL,HIGH);
    digitalWrite(CALC_EQUALS,LOW);
    digitalWrite(BONUS_LIGHT,LOW);
    delay(d/2);    
    digitalWrite(BELL,LOW);  
    delay(d/2); 
  }
}

void take_it_easy_on(){                         //take it easy alarm ON
int t=millis()/500;
if (t%2){
    digitalWrite(TIE_LIGHT,HIGH);
    digitalWrite(BUZZER,HIGH);
  }else{
    digitalWrite(TIE_LIGHT,LOW);
    digitalWrite(BUZZER,LOW);
  }
}

void take_it_easy_off(){                         //take it easy alarm OFF
  digitalWrite(TIE_LIGHT,LOW);
  digitalWrite(BUZZER,LOW);
}
